# Source from or copy to ~/.bashrc
function __complete_runcommands {
    local candidates=$(runcommands-complete -- "$COMP_LINE" "${COMP_WORDS[COMP_CWORD]}" $COMP_POINT bash)
    COMPREPLY=( $(compgen -W "${candidates}" -- $2) )
}

complete -F __complete_runcommands run runcommand runcommands commands.py
